/*----------------------------------------------------------------------
  File    : coco4r.c
  Contents: COntinuous time ClOsed Neuron Assembly Detection for R
  Author  : Christian Borgelt
  History : 2015.08.05 file created
            2015.08.10 pattern set reduction added
            2015.08.13 function c4r_patspec() renamed to c4r_genpsp()
            2015.08.13 adapted to generic function genpsp()
            2015.08.19 adapted to generic pattern set reduction
            2016.04.06 patterns changed to item set/support pairs
            2016.10.03 adapted to extensions of coconad implementation
            2016.10.21 adapted to interface changes of coconad impl.
            2016.10.23 closed/maximal repository coded in operation mode
            2016.11.18 coconad miner object and interface introduced
            2019.05.13 R_registerRoutines/R_useDynamicSymbols added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#include <R.h>
#include <Rinternals.h>
#ifndef TRN_SURR
#define TRN_SURR
#endif
#ifndef PSP_ESTIM
#define PSP_ESTIM
#endif
#ifndef PSP_TRAIN
#define PSP_TRAIN
#endif
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#ifndef ISR_CLOMAX
#define ISR_CLOMAX
#endif
#include "sigint.h"
#include "random.h"
#include "coconad.h"
#include "ccnpsp.h"
#include "patred.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE         1024    /* block size for arrays */
#define IS_NA(x)        (isnan(x) || isinf(x) || ((x) < 0))

/* --- error handling --- */
#define MYERROR(msg)    do { sig_remove(); error(msg); } while (0)
#define ERR_MEM()       MYERROR("out of memory")
#define ERR_ABORT()     MYERROR("user abort")

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- item set report data --- */
  SEXP   res;                   /* constructed result */
  size_t size;                  /* size of frequent item set arrays */
  size_t cnt;                   /* number of frequent item sets */
  int    err;                   /* error flag */
} REPDATA;                      /* (item set report data) */

/*----------------------------------------------------------------------
  Parameter Functions
----------------------------------------------------------------------*/

static int get_lgl (SEXP p, int dflt)
{                               /* --- get a logical parameter */
  assert(p);                    /* check the function argument */
  if (length(p) <  1)       return dflt;
  if (TYPEOF(p) == LGLSXP)  return (LOGICAL(p)[0])      ? -1 : 0;
  if (TYPEOF(p) == INTSXP)  return (INTEGER(p)[0] != 0) ? -1 : 0;
  if (TYPEOF(p) == REALSXP) return (REAL(p)[0]    != 0) ? -1 : 0;
  return dflt;                  /* extract and convert first element */
}  /* get_lgl() */

/*--------------------------------------------------------------------*/

static int get_int (SEXP p, int dflt)
{                               /* --- get an integer parameter */
  assert(p);                    /* check the function argument */
  if (length(p) <  1)       return dflt;
  if (TYPEOF(p) == LGLSXP)  return (LOGICAL(p)[0]) ? 1 : 0;
  if (TYPEOF(p) == INTSXP)  return (int)INTEGER(p)[0];
  if (TYPEOF(p) == REALSXP) return (int)REAL(p)[0];
  return dflt;                  /* extract and convert first element */
}  /* get_int() */

/*--------------------------------------------------------------------*/

static long int get_lng (SEXP p, long int dflt)
{                               /* --- get an integer parameter */
  assert(p);                    /* check the function argument */
  if (length(p) <  1)       return dflt;
  if (TYPEOF(p) == LGLSXP)  return (LOGICAL(p)[0]) ? 1 : 0;
  if (TYPEOF(p) == INTSXP)  return (long int)INTEGER(p)[0];
  if (TYPEOF(p) == REALSXP) return (long int)REAL(p)[0];
  return dflt;                  /* extract and convert first element */
}  /* get_lng() */

/*--------------------------------------------------------------------*/

static double get_dbl (SEXP p, double dflt)
{                               /* --- get a double parameter */
  assert(p);                    /* check the function argument */
  if (length(p) <  1)       return dflt;
  if (TYPEOF(p) == LGLSXP)  return (LOGICAL(p)[0]) ? 1.0 : 0.0;
  if (TYPEOF(p) == INTSXP)  return (double)INTEGER(p)[0];
  if (TYPEOF(p) == REALSXP) return (double)REAL(p)[0];
  return dflt;                  /* extract and convert first element */
}  /* get_dbl() */

/*--------------------------------------------------------------------*/

static CCHAR* get_str (SEXP p, CCHAR *dflt)
{                               /* --- get a string parameter */
  assert(p);                    /* check the function argument */
  if (length(p) <  1)       return dflt;
  if (TYPEOF(p) == STRSXP)  return CHAR(STRING_ELT(p, 0));
  return dflt;                  /* extract and convert first element */
}  /* get_str() */

/*--------------------------------------------------------------------*/

static int get_target (SEXP p)
{                               /* --- get target */
  CCHAR *s;                     /* target string */
  int   t;                      /* sequence flag */

  assert(p);                    /* check the function argument */
  if (TYPEOF(p) != STRSXP) return -1;
  if (length(p) < 1) s = "s";   /* get the target string */
  else s = CHAR(STRING_ELT(p, 0));
  if (*s && !s[1] && !s[2]) {   /* evaluate textual identifier */
    if      (strcmp(s, "set")          == 0) s = "fi";
    else if (strcmp(s, "sets")         == 0) s = "fi";
    else if (strcmp(s, "itemset")      == 0) s = "fi";
    else if (strcmp(s, "itemsets")     == 0) s = "fi";
    else if (strcmp(s, "all")          == 0) s = "fi";
    else if (strcmp(s, "allset")       == 0) s = "fi";
    else if (strcmp(s, "allsets")      == 0) s = "fi";
    else if (strcmp(s, "frq")          == 0) s = "fi";
    else if (strcmp(s, "freq")         == 0) s = "fi";
    else if (strcmp(s, "frequent")     == 0) s = "fi";
    else if (strcmp(s, "frqset")       == 0) s = "fi";
    else if (strcmp(s, "frqsets")      == 0) s = "fi";
    else if (strcmp(s, "freqset")      == 0) s = "fi";
    else if (strcmp(s, "freqsets")     == 0) s = "fi";
    else if (strcmp(s, "cls")          == 0) s = "ci";
    else if (strcmp(s, "clsd")         == 0) s = "ci";
    else if (strcmp(s, "closed")       == 0) s = "ci";
    else if (strcmp(s, "max")          == 0) s = "mi";
    else if (strcmp(s, "maxi")         == 0) s = "mi";
    else if (strcmp(s, "maximal")      == 0) s = "mi";
    else if (strcmp(s, "perm")         == 0) s = "fp";
    else if (strcmp(s, "perms")        == 0) s = "fp";
    else if (strcmp(s, "permutation")  == 0) s = "fp";
    else if (strcmp(s, "permutations") == 0) s = "fp";
    else if (strcmp(s, "allperm")      == 0) s = "fp";
    else if (strcmp(s, "allperms")     == 0) s = "fp";
    else if (strcmp(s, "frqperm")      == 0) s = "fp";
    else if (strcmp(s, "frqperms")     == 0) s = "fp";
    else if (strcmp(s, "freqperm")     == 0) s = "fp";
    else if (strcmp(s, "freqperms")    == 0) s = "fp";
    else if (strcmp(s, "clsperm")      == 0) s = "cp";
    else if (strcmp(s, "clsperms")     == 0) s = "cp";
    else if (strcmp(s, "clsdperm")     == 0) s = "cp";
    else if (strcmp(s, "clsdperms")    == 0) s = "cp";
    else if (strcmp(s, "closedperm")   == 0) s = "cp";
    else if (strcmp(s, "closedperms")  == 0) s = "cp";
    else if (strcmp(s, "maxperm")      == 0) s = "mp";
    else if (strcmp(s, "maxperms")     == 0) s = "mp";
    else if (strcmp(s, "maxiperm")     == 0) s = "mp";
    else if (strcmp(s, "maxiperms")    == 0) s = "mp";
    else if (strcmp(s, "maximalperm")  == 0) s = "mp";
    else if (strcmp(s, "maximalperms") == 0) s = "mp";
    else if (strcmp(s, "seq")          == 0) s = "fs";
    else if (strcmp(s, "seqs")         == 0) s = "fs";
    else if (strcmp(s, "sequence")     == 0) s = "fs";
    else if (strcmp(s, "sequences")    == 0) s = "fs";
    else if (strcmp(s, "allseq")       == 0) s = "fs";
    else if (strcmp(s, "allseqs")      == 0) s = "fs";
    else if (strcmp(s, "frqseq")       == 0) s = "fs";
    else if (strcmp(s, "frqseqs")      == 0) s = "fs";
    else if (strcmp(s, "freqseq")      == 0) s = "fs";
    else if (strcmp(s, "freqseqs")     == 0) s = "fs";
    else if (strcmp(s, "clsseq")       == 0) s = "cs";
    else if (strcmp(s, "clsseqs")      == 0) s = "cs";
    else if (strcmp(s, "clsdseq")      == 0) s = "cs";
    else if (strcmp(s, "clsdseqs")     == 0) s = "cs";
    else if (strcmp(s, "closedseq")    == 0) s = "cs";
    else if (strcmp(s, "closedseqs")   == 0) s = "cs";
    else if (strcmp(s, "maxseq")       == 0) s = "ms";
    else if (strcmp(s, "maxseqs")      == 0) s = "ms";
    else if (strcmp(s, "maxiseq")      == 0) s = "ms";
    else if (strcmp(s, "maxiseqs")     == 0) s = "ms";
    else if (strcmp(s, "maximalseq")   == 0) s = "ms";
    else if (strcmp(s, "maximalseqs")  == 0) s = "ms";
  }
  if (*s && (!s[1] || !s[2])) { /* check for a valid string */
    switch (s[1]) {             /* evaluate the sequence indicator */
      case  0 : t = CCN_ITEMSET;  break;
      case 'i': t = CCN_ITEMSET;  break;
      case 'p': t = CCN_PERMUT;   break;
      case 'o': t = CCN_PERMUT;   break;
      case 's': t = CCN_SEQUENCE; break;
      case 'q': t = CCN_SEQUENCE; break;
      default : t = -1;           break;
    }
    switch (s[0]) {             /* evaluate the target code */
      case 's': return t|ISR_FREQUENT;
      case 'a': return t|ISR_FREQUENT;
      case 'f': return t|ISR_FREQUENT;
      case 'c': return t|ISR_CLOSED;
      case 'm': return t|ISR_MAXIMAL;
    }
  }
  return -1;                    /* return an error code */
}  /* get_target() */

/*--------------------------------------------------------------------*/

static int get_coconad (SEXP p, int target)
{                               /* --- get CoCoNAD algorithm */
  CCHAR *s;                     /* algorithm string */

  assert(p);                    /* check the function argument */
  if (TYPEOF(p) != STRSXP) return -1;
  if (length(p) < 1) s = "";    /* get the algorithm string */
  else s = CHAR(STRING_ELT(p, 0));
  if (!*s) s = (target & (CCN_PERMUT|CCN_SEQUENCE)) ? "f" : "r";
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "basic")      == 0) s = "b";
    else if (strcmp(s, "base")       == 0) s = "b";
    else if (strcmp(s, "filt")       == 0) s = "f";
    else if (strcmp(s, "filter")     == 0) s = "f";
    else if (strcmp(s, "filtered")   == 0) s = "f";
    else if (strcmp(s, "refined")    == 0) s = "f";
    else if (strcmp(s, "repo")       == 0) s = "r";
    else if (strcmp(s, "repository") == 0) s = "r";
    else if (strcmp(s, "cmrepo")     == 0) s = "r";
  }
  if (s[0] && !s[1]) {          /* check for a valid string */
    switch (s[0]) {             /* evaluate the algorithm code */
      case 'b': return CCN_BASIC;
      case 'f': return CCN_FILTER;
      case 'r': return CCN_REPO;
    }
  }
  return -1;                    /* return an error code */
}  /* get_coconad() */

/*--------------------------------------------------------------------*/

static int get_surr (SEXP p)
{                               /* --- get surrogate function code */
  CCHAR *s;                     /* surrogate string */

  assert(p);                    /* check the function argument */
  if (TYPEOF(p) != STRSXP) return -1;
  if (length(p) < 1) s = "p";   /* get the surrogate function string */
  else s = CHAR(STRING_ELT(p, 0));
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "ident")     == 0) s = "i";
    else if (strcmp(s, "identity")  == 0) s = "i";
    else if (strcmp(s, "random")    == 0) s = "r";
    else if (strcmp(s, "randomize") == 0) s = "r";
    else if (strcmp(s, "dither")    == 0) s = "d";
    else if (strcmp(s, "displace")  == 0) s = "d";
    else if (strcmp(s, "jitter")    == 0) s = "d";
    else if (strcmp(s, "shift")     == 0) s = "s";
    else if (strcmp(s, "kernel")    == 0) s = "k";
    else if (strcmp(s, "kernest")   == 0) s = "k";
    else if (strcmp(s, "perm")      == 0) s = "p";
    else if (strcmp(s, "permute")   == 0) s = "p";
    else if (strcmp(s, "block")     == 0) s = "b";
    else if (strcmp(s, "blocked")   == 0) s = "b";
    else if (strcmp(s, "blkperm")   == 0) s = "b";
    else if (strcmp(s, "blockperm") == 0) s = "b";
  }
  if (s[0] && !s[1]) {          /* translate surrogate method string */
    switch (s[0]) {             /* evaluate the surrogate method code */
      case 'i': return CCN_IDENTITY;
      case 'r': return CCN_RANDOM;
      case 'd': return CCN_DITHER;
      case 's': return CCN_SHIFT;
      case 'k': return CCN_KERNEST;
      case 'p': return CCN_PERMUTE;
      case 'b': return CCN_BLOCK;
    }
  }
  return -1;                    /* return an error code */
}  /* get_surr() */

/*--------------------------------------------------------------------*/

static int get_rand (SEXP p)
{                               /* --- get random function code */
  CCHAR *s;                     /* density function string */

  assert(p);                    /* check the function argument */
  if (TYPEOF(p) != STRSXP) return -1;
  if (length(p) < 1) s = "u";   /* get the random function string */
  else s = CHAR(STRING_ELT(p, 0));
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "uni")         == 0) s = "u";
    else if (strcmp(s, "uniform")     == 0) s = "u";
    else if (strcmp(s, "rect")        == 0) s = "r";
    else if (strcmp(s, "rectangle")   == 0) s = "r";
    else if (strcmp(s, "rectangular") == 0) s = "r";
    else if (strcmp(s, "triang")      == 0) s = "t";
    else if (strcmp(s, "triangle")    == 0) s = "t";
    else if (strcmp(s, "triangular")  == 0) s = "t";
    else if (strcmp(s, "Gauss")       == 0) s = "g";
    else if (strcmp(s, "Gaussian")    == 0) s = "g";
    else if (strcmp(s, "gauss")       == 0) s = "g";
    else if (strcmp(s, "gaussian")    == 0) s = "g";
    else if (strcmp(s, "normal")      == 0) s = "n";
  }
  if (s[0] && !s[1]) {          /* translate surrogate method string */
    switch (s[0]) {             /* evaluate the surrogate method code */
      case 'u': return RNG_UNIFORM;
      case 'r': return RNG_RECT;
      case 't': return RNG_TRIANG;
      case 'g': return RNG_GAUSS;
      case 'n': return RNG_NORMAL;
    }
  }
  return -1;                    /* return an error code */
}  /* get_rand() */

/*--------------------------------------------------------------------*/

static int get_red (SEXP p)
{                               /* --- get random function code */
  CCHAR *s;                     /* density function string */

  assert(p);                    /* check the function argument */
  if (TYPEOF(p) != STRSXP) return -1;
  if (length(p) < 1) s = "S";   /* get the random function string */
  else s = CHAR(STRING_ELT(p, 0));
  if (s[0] && s[1]) {           /* evaluate textual identifier */
    if      (strcmp(s, "none")        == 0) s = "x";
    else if (strcmp(s, "coins")       == 0) s = "c";
    else if (strcmp(s, "coins0")      == 0) s = "c";
    else if (strcmp(s, "coins1")      == 0) s = "C";
    else if (strcmp(s, "coins+1")     == 0) s = "C";
    else if (strcmp(s, "items")       == 0) s = "i";
    else if (strcmp(s, "items2")      == 0) s = "i";
    else if (strcmp(s, "neurons")     == 0) s = "i";
    else if (strcmp(s, "cover")       == 0) s = "s";
    else if (strcmp(s, "cover0")      == 0) s = "s";
    else if (strcmp(s, "covered")     == 0) s = "s";
    else if (strcmp(s, "covered0")    == 0) s = "s";
    else if (strcmp(s, "cover1")      == 0) s = "S";
    else if (strcmp(s, "covered1")    == 0) s = "S";
    else if (strcmp(s, "leni")        == 0) s = "l";
    else if (strcmp(s, "leni0")       == 0) s = "l";
    else if (strcmp(s, "lenient")     == 0) s = "l";
    else if (strcmp(s, "lenient0")    == 0) s = "l";
    else if (strcmp(s, "leni1")       == 0) s = "L";
    else if (strcmp(s, "lenient1")    == 0) s = "L";
    else if (strcmp(s, "strict")      == 0) s = "t";
    else if (strcmp(s, "strict0")     == 0) s = "t";
    else if (strcmp(s, "strict1")     == 0) s = "T";
  }
  if (s[0] && !s[1]) {          /* translate surrogate method string */
    switch (s[0]) {             /* evaluate the surrogate method code */
      case 'x': return PSR_NONE;
      case 'c': return PSR_COINS0;
      case 'C': return PSR_COINS1;
      case 'i': return PSR_ITEMS2;
      case 's': return PSR_COVER0;
      case 'S': return PSR_COVER1;
      case 'l': return PSR_LENIENT0;
      case 'L': return PSR_LENIENT1;
      case 't': return PSR_STRICT0;
      case 'T': return PSR_STRICT1;
    }
  }
  return -1;                    /* return an error code */
}  /* get_red() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static int chktrains (SEXP trains)
{                               /* --- check trains argument */
  R_xlen_t i, n;                /* loop variables */
  int      t;                   /* type of the trains */

  assert(trains);               /* check the function arguments */
  if (TYPEOF(trains) != VECSXP) return -1;
  n = length(trains);           /* check the data type */
  if (n <= 0) return -1;        /* and the list length */
  for (i = 0; i < n; i++) {     /* traverse the list elements */
    t = TYPEOF(VECTOR_ELT(trains, i));
    if ((t != INTSXP) && (t != REALSXP)) return -2;
  }                             /* check for a numeric vector */
  return 0;                     /* return 'ok' */
}  /* chktrains() */

/*--------------------------------------------------------------------*/

static TRAINSET* tns_fromRObj (SEXP trains)
{                               /* --- create a train set */
  int      e = 0;               /* error flag */
  ITEM     i, n;                /* item identifier, loop variable */
  TID      k, m;                /* array length,    loop variable */
  SEXP     p;                   /* to traverse the trains */
  int      *src;                /* source array for integers */
  ITEMBASE *ibase;              /* underlying item base */
  TRAINSET *trnset;             /* created train set */
  TRAIN    *trn;                /* to create the trains */

  assert(trains);               /* check the function argument */
  ibase = ib_create(IB_OBJNAMES, 0, ST_INTFN, (OBJFN*)0);
  if (!ibase) return NULL;      /* create an item base */
  n = (ITEM)length(trains);     /* get the number of trains */
  for (i = 1; i <= n; i++)      /* add all item names (integers) */
    if (ib_add(ibase, &i) < 0) break;
  if (i < n) { ib_delete(ibase); return NULL; }
  trnset = tns_create(ibase);   /* create a (spike) train set */
  if (!trnset) { ib_delete(ibase); return NULL; }
  for (i = 0; i < n; i++) {     /* traverse the (spike) trains */
    p = VECTOR_ELT(trains, i);  /* get the next (spike) train */
    m = (TID)length(p);         /* and its length */
    if      (TYPEOF(p) == REALSXP)  /* if floating point array */
      trn = trn_create(i, REAL(p), m);
    else if (TYPEOF(p) == INTSXP) { /* if integer array */
      src = INTEGER(p);         /* create an empty train */
      trn = trn_create(i, NULL,    m);
      for (k = 0; k < m; k++) trn_add(trn, (double)src[k]); }
    else trn = NULL;            /* copy integers one by one */
    if      (!trn) e = -1;      /* create a (spike) train object */
    else if (tns_add(trnset, trn) < 0) {
      e = -1; trn_delete(trn); }/* add created (spike) train to set */
    if (e) break;               /* abort the loop */
  }                             /* if an error occurred */
  if (e) { tns_delete(trnset, 1); return NULL; }
  return trnset;                /* return the created train set */
}  /* tns_fromRObj() */

/*--------------------------------------------------------------------*/

static int isr_Rborder (ISREPORT *rep, SEXP border)
{                               /* --- set reporter filtering border */
  ITEM   n;                     /* loop variables, border length */
  RSUPP  s;                     /* minimum support */
  int    *b;                    /* elements of the border array */
  double *d;                    /* elements of the border array */

  assert(rep && border);        /* check the function arguments */
  if (border == R_NilValue) return 0;
  n = (ITEM)length(border);     /* if no border is given or */
  if (n <= 0) return 0;         /* if it has length 0, abort */
  if      (TYPEOF(border) == INTSXP) {
    b = INTEGER(border);        /* if integer border */
    while (--n >= 0) {          /* traverse the border elements */
      s = (b[n] < 0)    ? RSUPP_MAX : (RSUPP)b[n];
      if (isr_setbdr(rep, (ITEM)n, s) < 0) return -1;
    } }                         /* set the support values */
  else if (TYPEOF(border) == REALSXP) {
    d = REAL(border);           /* if real-valued border */
    while (--n >= 0) {          /* traverse the border elements */
      s = (IS_NA(d[n])) ? RSUPP_MAX : (RSUPP)d[n];
      if (isr_setbdr(rep, (ITEM)n, s) < 0) return -1;
    }                           /* set the support values */
  }                             /* in the item set reporter */
  return 0;                     /* return 'ok' */
}  /* isr_Rborder() */

/*--------------------------------------------------------------------*/

static void isr_iset2RObj (ISREPORT *rep, void *data)
{                               /* --- report an item set */
  REPDATA *rd = data;           /* type the data pointer */
  size_t  k, n;                 /* loop variable, array size */
  ITEM    i, m;                 /* loop variable, array size */
  SEXP    p;                    /* resized R vector/list */
  SEXP    rset;                 /* new R item set */
  int     *iset;                /* item set elements */
  SEXP    info;                 /* information for item set */
  SEXP    relt;                 /* result element */

  assert(rep && data);          /* check the function arguments */
  if (rd->err) return;          /* if there was an error, do nothing */
  n = rd->size;                 /* get the current array size */
  if (rd->cnt >= n) {           /* if the item set array is full */
    n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
    p = PROTECT(allocVector(VECSXP, (R_xlen_t)n));
    if (rd->res) {              /* if there is an old vector/list */
      for (k = 0; k < rd->cnt; k++)
        SET_VECTOR_ELT(p, (R_xlen_t)k, VECTOR_ELT(rd->res,(R_xlen_t)k));
      UNPROTECT(2); PROTECT(p); /* copy the existing item sets and */
    }                           /* transfer protection to new array */
    rd->res = p; rd->size = n;  /* set the (new) vector/list */
  }                             /* and the new array size */
  m    = isr_cnt(rep);          /* create object for item set */
  rset = PROTECT(allocVector(INTSXP, (R_xlen_t)m));
  iset = INTEGER(rset);         /* get the integer elements */
  for (i = 0; i < m; i++)       /* map identifiers to items */
    iset[i] = (int)(ptrdiff_t)isr_itemobj(rep, isr_itemx(rep, i));
  info = PROTECT(allocVector(REALSXP, 1)); /* get and store */
  REAL(info)[0] = (double) isr_supp(rep);  /* item set support */
  relt = PROTECT(allocVector(VECSXP, 2));
  SET_VECTOR_ELT(relt,0,rset);  /* build element for each item set */
  SET_VECTOR_ELT(relt,1,info);  /* and store set and information */
  SET_VECTOR_ELT(rd->res, (R_xlen_t)rd->cnt, relt);
  rd->cnt += 1;                 /* store and count created item set */
  UNPROTECT(3);                 /* release the sub-objects */
}  /* isr_iset2RObj() */

/*--------------------------------------------------------------------*/

static SEXP psp_toRObj (PATSPEC *psp, double scale, int format)
{                               /* --- report pattern spectrum */
  ITEM   z;                     /* loop variable for sizes */
  SUPP   s;                     /* loop variable for support values */
  size_t i, n;                  /* list size, result list index */
  size_t k;                     /* number of occurrences */
  double *r;                    /* to store the values */
  SEXP   res;                   /* created R list object */
  SEXP   row;                   /* R object  for row-wise    rep. */
  SEXP   col[3];                /* R objects for column-wise rep. */
  int    *sizes;                /* pattern sizes */
  int    *supps;                /* support values */
  double *freqs;                /* occurrence frequencies */

  assert(psp);                  /* check the function arguments */
  n = psp_sigcnt(psp);          /* get the number of signatures */
  if ((format == '=')           /* if list/vector of triplets */
  ||  (format == '-')) {        /* (size, support, frequency) */
    res = PROTECT(allocVector(VECSXP, (R_xlen_t)n));
    for (i = 0, z = psp_min(psp); z <= psp_max(psp); z++) {
      for (s = psp_min4sz(psp, z); s <= psp_max4sz(psp, z); s++) {
        k = psp_getfrq(psp,z,s);/* traverse the pattern signatures */
        if (k <= 0) continue;   /* skip empty counters */
        row  = PROTECT(allocVector(REALSXP, 3));
        r    = REAL(row);       /* create a row with three elements */
        r[0] = (double)z;       /* store the patter size, */
        r[1] = (double)s;       /* the pattern support, */
        r[2] = (double)k *scale;/* and the frequency */
        SET_VECTOR_ELT(res, (R_xlen_t)i, row); i += 1;
        UNPROTECT(1);           /* set the row in the output vector */
      }                         /* and count and unprotect the row */
    } }                         /* (protected as part of 'res') */
  else {                        /* if three arrays (size,supp,freq) */
    res    = PROTECT(allocVector(VECSXP,  (R_xlen_t)3));
    col[0] = PROTECT(allocVector(INTSXP,  (R_xlen_t)n));
    SET_VECTOR_ELT(res, 0, col[0]); UNPROTECT(1);
    col[1] = PROTECT(allocVector(INTSXP,  (R_xlen_t)n));
    SET_VECTOR_ELT(res, 1, col[1]); UNPROTECT(1);
    col[2] = PROTECT(allocVector(REALSXP, (R_xlen_t)n));
    SET_VECTOR_ELT(res, 2, col[2]); UNPROTECT(1);
    sizes = INTEGER(col[0]);    /* create two integer and */
    supps = INTEGER(col[1]);    /* one real-valued array for */
    freqs = REAL   (col[2]);    /* size, support and frequency */
    for (i = 0, z = psp_min(psp); z <= psp_max(psp); z++) {
      for (s = psp_min4sz(psp, z); s <= psp_max4sz(psp, z); s++) {
        k = psp_getfrq(psp,z,s);/* traverse the pattern signatures */
        if (k <= 0) continue;   /* skip empty counters */
        sizes[i]   = z; supps[i] = s;
        freqs[i++] = (double)k *scale;
      }                         /* store the signature elements */
    }                           /* in the corresponding arrays */
  }
  assert(i == n);               /* check signature counter */
  return res;                   /* return created pattern spectrum */
}  /* psp_toRObj() */

/*--------------------------------------------------------------------*/

static int repinit (REPDATA *data, ISREPORT *isrep, CCHAR *report)
{                               /* --- initialize reporting */
  assert(data && isrep && report); /* check the function arguments */
  data->err = 0;                /* initialize the error indicator */
  if ((report[0] == '#')        /* if to get a pattern spectrum */
  ||  (report[0] == '|')        /* "#", "|": column-wise */
  ||  (report[0] == '=')        /* "=", "-": row-wise */
  ||  (report[0] == '-'))
    return isr_addpsp(isrep, NULL);
  data->res   = NULL;           /* initialize the report structure */
  data->size  = data->cnt = 0;  /* and set the reporting function */
  isr_setrepo(isrep, isr_iset2RObj, data);
  return 0;                     /* return 'ok' */
}  /* repinit() */

/*--------------------------------------------------------------------*/

static int repterm (REPDATA *data, ISREPORT *isrep, CCHAR *report)
{                               /* --- terminate reporting */
  size_t k;                     /* loop variable */
  SEXP   p;                     /* resized R vector/list */

  assert(data && isrep && report); /* check the function arguments */
  if (data->err) return data->err -1;
  if ((report[0] == '#')        /* if to get a pattern spectrum */
  ||  (report[0] == '|')        /* "#", "|": column-wise */
  ||  (report[0] == '=')        /* "=", "-": row-wise */
  ||  (report[0] == '-')) {
    data->res = psp_toRObj(isr_getpsp(isrep), 1.0, report[0]);
    return data->err = (data->res) ? 0 : -1;
  }                             /* make R pattern spectrum */
  if (data->cnt != data->size) {/* if result list has wrong size */
    p = PROTECT(allocVector(VECSXP, (R_xlen_t)data->cnt));
    for (k = 0; k < data->cnt; k++) /* create a new result list/array */
      SET_VECTOR_ELT(p, (R_xlen_t)k, VECTOR_ELT(data->res,(R_xlen_t)k));
    UNPROTECT(2); PROTECT(p);   /* copy the existing item sets */
    data->res = p; data->size = data->cnt;
  }                             /* set array/list and its size */
  return data->err;             /* return the error status */
}  /* repterm() */

/*--------------------------------------------------------------------*/

static void repfn (long int cnt, void *data)
{                               /* --- progress reporting function */
#if 0
  if ((cnt > *(long int*)data) && ((cnt % 20) == 0))
    fprintf(stderr, "%10ld\b\b\b\b\b\b\b\b\b\b",
                    *(long int*)data = cnt);
#endif
}  /* repfn() */

/*--------------------------------------------------------------------*/
/* coconad (trains, target="c", width=0.003, supp=2,                  */
/*          zmin=2, zmax=-1, report="a", algo="r", mode="",           */
/*          border=NULL)                                              */
/*--------------------------------------------------------------------*/

SEXP c4r_coconad (SEXP ptrains, SEXP ptarget, SEXP pwidth, SEXP psupp,
                  SEXP pzmin, SEXP pzmax, SEXP preport,
                  SEXP palgo, SEXP pmode, SEXP pborder)
{                               /* --- CoCoNAD algorithm */
  int      target  = ISR_CLOSED;/* target type identifier */
  double   width   =  0.003;    /* width of influence region */
  SUPP     supp    =  2;        /* minimum support of an item set */
  int      zmin    =  2;        /* minimum size of an item set */
  int      zmax    = -1;        /* maximum size of an item set */
  CCHAR    *report = "a";       /* indicators of values to report */
  int      algo    = CCN_REPO;  /* CoCoNAD algorithm */
  int      mode    = CCN_DEFAULT;  /* operation mode */
  int      mrep    =  0;        /* reporting mode */
  TRAINSET *trnset;             /* set of (spike) trains */
  ITEM     n;                   /* number of (spike) trains */
  ISREPORT *isrep;              /* item set reporter */
  COCONAD  *coconad;            /* coconad miner */
  REPDATA  data;                /* data for item set reporting */
  CCHAR    *s;                  /* to traverse the mode flags */
  int      r;                   /* result of function call */

  /* --- evaluate function arguments --- */
  r = chktrains(ptrains);       /* check the given trains */
  if (r < -1) error("invalid 'trains' argument "
                    "(must be list of numeric arrays)");
  if (r <  0) error("invalid 'trains' argument "
                    "(must be non-empty list)");
  target = get_target(ptarget); /* translate the target string */
  if (target < 0)    error("invalid 'target' argument");
  width  = get_dbl(pwidth, width);
  if (width  < 0)    error("invalid 'width' argument (must be >= 0)");
  supp   = (SUPP)get_int(psupp, (int)supp);
  if (supp   < 0)    supp = -supp; /* ensure non-negative support */
  zmin   = get_int(pzmin,  zmin);
  if (zmin   < 1)    error("invalid 'zmin' argument (must be > 0)");
  zmax   = get_int(pzmax,  zmax);
  if (zmax   < 1)    zmax = ITEM_MAX;  /* check the size range */
  if (zmax   < zmin) error("invalid 'zmax' argument (must be >= zmin)");
  report = get_str(preport, report);
  algo   = get_coconad(palgo, target); /* translate algorithm code */
  if (algo   < 0)    error("invalid 'algo' argument");
  for (s = get_str(pmode, ""); *s; s++) {
    if      (*s == 'x') mode &= ~CCN_PERFECT;
    else if (*s == 'p') mode &= ~CCN_REORDER;
    else if (*s == '1') mode |=  CCN_REPOTREE;
    else if (*s == '0') mode |=  CCN_REPOLIST;
    else error("invalid 'mode' argument");
  }                             /* adapt the operation mode */
  if (algo != CCN_REPO) mrep |= ISR_NOFILTER;
  if ((pborder != R_NilValue)   /* check the filtering border */
  &&  (TYPEOF(pborder) != INTSXP) && (TYPEOF(pborder) != REALSXP))
    error("invalid 'border' argument (must be numeric)");

  /* --- create train set --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromRObj(ptrains);
  if (!trnset) ERR_MEM();       /* create and initialize a train set */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  coconad = ccn_create(target, width, supp,
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
                       mix,     /* add. support mixing parameter */
  #endif
                       (ITEM)zmin, (ITEM)zmax, algo, mode);
  if (!coconad) {               /* create a coconad miner */
    tns_delete(trnset,  1); ERR_MEM(); }
  if (ccn_data(coconad, trnset, 0) != 0) {
    ccn_delete(coconad, 1); ERR_MEM(); }

  /* --- create item set reporter --- */
  n = tns_cnt(trnset);          /* get the number of trains */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (target & CCN_SEQUENCE)    /* if sequences (with poss. reps.) */
    n = tns_maxseq(trnset, width) +1;
  #endif                        /* find max. possible sequence len. */
  isrep = isr_createx(tns_base(trnset), n);
  if (!isrep                    /* create an item set reporter */
  ||  (ccn_report(coconad, isrep)    != 0)
  ||  (isr_Rborder(isrep, pborder)   != 0)
  ||  (repinit(&data, isrep, report) != 0)
  ||  (isr_setup(isrep) < 0)) { /* prepare the item set reporter */
    ccn_delete(coconad, 1); ERR_MEM(); }

  /* --- frequent item set mining --- */
  r = ccn_mine(coconad);
  if (r >= 0) r = repterm(&data, isrep, report);

  /* --- clean up --- */
  ccn_delete(coconad, 1);       /* delete the coconad miner */
  if (data.res) UNPROTECT(1);   /* unprotect the result object */
  if (sig_aborted()) { sig_abort(0); ERR_ABORT(); }
  sig_remove();                 /* remove the signal handler */
  if (r != 0) ERR_MEM();        /* check for an error */
  return (data.res) ? data.res : allocVector(VECSXP, 0);
}  /* c4r_coconad() */          /* return the created R object */

/*--------------------------------------------------------------------*/
/* genpsp (trains, target="c", width=0.003, supp=2,                   */
/*         zmin=2, zmax=-1, report="|", algo="r", mode="",            */
/*         cnt=1000, beg=-Inf, end=+Inf, surr="p", rand="u",          */
/*         sigma=0.005, delta=0.03, seed=0, cpus=0)                   */
/*--------------------------------------------------------------------*/

SEXP c4r_genpsp (SEXP ptrains, SEXP ptarget, SEXP pwidth,
                 SEXP psupp,   SEXP pzmin, SEXP pzmax,
                 SEXP preport, SEXP palgo, SEXP pmode,
                 SEXP pcnt,    SEXP pbeg,  SEXP pend,
                 SEXP psurr,   SEXP prand, SEXP psigma, SEXP pdelta,
                 SEXP pseed,   SEXP pcpus)
{                               /* --- pattern spectrum finding */
  int      target  = ISR_CLOSED;/* target type identifier */
  double   width   =  0.003;    /* width of influence region */
  SUPP     supp    =  2;        /* minimum support of an item set */
  int      zmin    =  2;        /* minimum size of an item set */
  int      zmax    = -1;        /* maximum size of an item set */
  CCHAR    *report = "|";       /* indicators of reporting format */
  int      algo    = CCN_REPO;  /* CoCoNAD algorithm */
  int      mode    = CCN_DEFAULT;  /* operation mode */
  int      cnt     = 1000;      /* number of data sets to generate */
  double   beg     = -INFINITY; /* beginning of point range */
  double   end     =  INFINITY; /* end       of point range */
  int      surr    =  5;        /* surrogate method identifier */
  int      rand    = RNG_RECT;  /* random function identifier */
  double   sigma   =  0.005;    /* parameter of random function */
  double   delta   =  0.03;     /* limit for localized swaps */
  long int seed    =  0;        /* seed for random number generator */
  int      cpus    =  0;        /* number of cpus */
  PATSPEC  *psp    = NULL;      /* created pattern spectrum */
  SEXP     rpsp    = NULL;      /* created R pattern spectrum */
  long int done    = 0;         /* number of completed data sets */
  TRAINSET *trnset;             /* created train set */
  CCHAR    *s;                  /* to traverse the mode flags */
  int      r;                   /* result of function call */

  /* --- evaluate the function arguments --- */
  r = chktrains(ptrains);       /* check the given trains */
  if (r < -1) error("invalid 'trains' argument "
                    "(must be list of numeric arrays)");
  if (r <  0) error("invalid 'trains' argument "
                    "(must be non-empty list)");
  target = get_target(ptarget); /* translate the target string */
  if (target < 0)    error("invalid 'target' argument");
  width  = get_dbl(pwidth, width);
  if (width  < 0)    error("invalid 'width' argument (must be >= 0)");
  supp   = (SUPP)get_int(psupp, (int)supp);
  if (supp   < 0)    supp = -supp; /* ensure non-negative support */
  zmin   = get_int(pzmin,  zmin);
  if (zmin   < 1)    error("invalid 'zmin' argument (must be > 0)");
  zmax   = get_int(pzmax,  zmax);
  if (zmax   < 1)    zmax = ITEM_MAX;  /* check the size range */
  if (zmax   < zmin) error("invalid 'zmax' argument (must be >= zmin)");
  report = get_str(preport, report);
  for (s = get_str(pmode, ""); *s; s++) {
    if      (*s == 'x') mode &= ~CCN_PERFECT;
    else if (*s == 'p') mode &= ~CCN_REORDER;
    else if (*s == '1') mode |=  CCN_REPOTREE;
    else if (*s == '0') mode |=  CCN_REPOLIST;
  }                             /* adapt the operation mode */
  algo   = get_coconad(palgo, target); /* translate algorithm code */
  if (algo   < 0)    error("invalid 'algo' argument");
  cnt    = get_int(pcnt, cnt);
  if (cnt   <= 0)    cnt = 1;   /* check the number of data sets */
  beg    = get_dbl(pbeg, beg);
  end    = get_dbl(pend, end);
  surr   = get_surr(psurr);     /* translate the surrogate string */
  if (surr  <  0)    error("invalid 'surr' argument");
  if (surr  == 0)    cnt = 1;   /* only one surrogate for identity */
  rand   = get_rand(prand);     /* translate random function string */
  if (rand  <  0)    error("invalid 'rand' argument");
  sigma  = get_dbl(psigma, sigma);
  if (sigma <  0)    error("invalid 'sigma' argument (must be >= 0)");
  delta  = get_dbl(pdelta, delta);
  if (delta <= 0)    error("invalid 'delta' argument (must be > 0)");
  seed   = get_lng(pseed, seed);
  cpus   = get_int(pcpus, cpus);

  /* --- generate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromRObj(ptrains);
  if (!trnset) ERR_MEM();       /* create a train set */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  psp = ccn_genpsp(trnset, target, width, supp, (ITEM)zmin, (ITEM)zmax,
                   algo, mode, (size_t)cnt, beg, end, surr, rand,
                   sigma, delta, seed, cpus, repfn, &done);
  if (psp) { rpsp = psp_toRObj(psp, 1.0/(double)cnt, report[0]);
             psp_delete(psp); } /* generate a pattern spectrum */
  tns_delete(trnset, 1);        /* delete the train set */
  if (!rpsp) ERR_MEM();         /* check for an error, otherwise */
  else       UNPROTECT(1);      /* unprotect the created object */
  if (sig_aborted()) { sig_abort(0); ERR_ABORT(); }
  sig_remove();                 /* remove the signal handler */
  return rpsp;                  /* return the created R object */
}  /* c4r_genpsp() */

/*--------------------------------------------------------------------*/
/* estpsp (trains, target="a", width=0.003, supp=2,                   */
/*         zmin=2, zmax=-1, report="|", equiv=10000,                  */
/*         alpha=0.5, smpls=1000, seed=0)                             */
/*--------------------------------------------------------------------*/

SEXP c4r_estpsp (SEXP ptrains, SEXP ptarget, SEXP pwidth, SEXP psupp,
                 SEXP pzmin, SEXP pzmax, SEXP preport, SEXP pequiv,
                 SEXP palpha, SEXP psmpls, SEXP pseed)
{                               /* --- pattern spectrum estimation */
  int      target  = ISR_ALL;   /* target pattern type identifier */
  double   width   =  0.003;    /* width of influence region */
  SUPP     supp    =  2;        /* minimum support of an item set */
  int      zmin    =  2;        /* minimum size of an item set */
  int      zmax    = -1;        /* maximum size of an item set */
  CCHAR    *report = "|";       /* indicators of reporting format */
  int      equiv   = 10000;     /* equivalent number of surrogates */
  double   alpha   =  0.5;      /* probability dispersion factor */
  int      smpls   = 1000;      /* number of samples per set size */
  long int seed    =  0;        /* seed for random number generator */
  PATSPEC  *psp    = NULL;      /* created pattern spectrum */
  SEXP     rpsp    = NULL;      /* created R pattern spectrum */
  TRAINSET *trnset;             /* created train set */
  int      r;                   /* result of function call */

  /* --- evaluate function arguments --- */
  r = chktrains(ptrains);       /* check the given trains */
  if (r < -1) error("invalid 'trains' argument "
                    "(must be list of numeric arrays)");
  if (r <  0) error("invalid 'trains' argument "
                    "(must be non-empty list)");
  target = get_target(ptarget); /* translate the target string */
  if (target < 0)    error("invalid 'target' argument");
  width  = get_dbl(pwidth, width);
  if (width  < 0)    error("invalid 'width' argument (must be >= 0)");
  supp   = (SUPP)get_int(psupp, (int)supp);
  if (supp   < 0)    supp = -supp; /* ensure non-negative support */
  zmin   = get_int(pzmin,  zmin);
  if (zmin   < 1)    error("invalid 'zmin' argument (must be > 0)");
  zmax   = get_int(pzmax,  zmax);
  if (zmax   < 1)    zmax = ITEM_MAX;  /* check the size range */
  if (zmax   < zmin) error("invalid 'zmax' argument (must be >= zmin)");
  report = get_str(preport, report);
  equiv  = get_int(pequiv,  equiv);
  if (equiv <= 0)    equiv = 1; /* check the number of data sets */
  alpha  = get_dbl(palpha,  alpha);
  if (alpha <= 0)    error("invalid 'alpha' argument (must be > 0)");
  smpls  = get_int(psmpls,  smpls);
  if (smpls <= 0)    error("invalid 'smpls' argument (must be > 0)");
  seed   = get_lng(pseed, seed);

  /* --- estimate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromRObj(ptrains);
  if (!trnset) ERR_MEM();       /* create a train set and */
  tns_sort(trnset, +1);         /* sort the trains by points */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  psp = ccn_estpsp(trnset, target, width, supp, (ITEM)zmin, (ITEM)zmax,
                   (size_t)equiv, alpha, (size_t)smpls, seed);
  if (psp) { rpsp = psp_toRObj(psp, 1.0/(double)equiv, report[0]);
             psp_delete(psp); } /* generate a pattern spectrum */
  tns_delete(trnset, 1);        /* delete the train set */
  if (!rpsp) ERR_MEM();         /* check for an error, otherwise */
  else       UNPROTECT(1);      /* unprotect the created object */
  if (sig_aborted()) { sig_abort(0); ERR_ABORT(); }
  sig_remove();                 /* remove the signal handler */
  return rpsp;                  /* return the created R object */
}  /* c4r_estpsp() */

/*--------------------------------------------------------------------*/
/* patred (pats, method="S", border=NULL, addis=TRUE)                 */
/*--------------------------------------------------------------------*/

SEXP c4r_patred (SEXP ppats, SEXP pmethod, SEXP pborder, SEXP paddis)
{                               /* --- pattern set reduction */
  int    method = PSR_COVER1;   /* pattern set reduction method */
  int    addis  = -1;           /* whether to add pairwise isects. */
  PATSET *patset;               /* internal pattern set */
  size_t i, n;                  /* loop variable, number of patterns */
  size_t k, z;                  /* (maximal) size of a pattern */
  SEXP   rset;                  /* to traverse the R item sets */
  SEXP   items;                 /* to traverse the item sets */
  SEXP   supp;                  /* (vector of) support value(s) */
  int    t;                     /* type of support (integer or real) */
  RSUPP  s;                     /* support of item set or in border */
  int    *b;                    /* to access an integer border */
  double *d;                    /* to access a real-valued border */

  /* --- evaluate function arguments --- */
  if (ppats == R_NilValue) return ppats;
  if (TYPEOF(ppats) != VECSXP)  /* pattern set must be a vector/array */
    error("invalid 'pats' argument (must be a list)");
  n = (size_t)length(ppats);    /* get the number of patterns */
  if (n <= 0) return ppats;     /* and return if there are none */
  method = get_red(pmethod);    /* get and check the reduction method */
  if (method < 0) error("invalid 'method' argument");
  addis = get_lgl(paddis,addis);/* get flag for adding intersections */
  if ((pborder != R_NilValue)   /* check the filtering border */
  &&  (TYPEOF(pborder) != INTSXP) && (TYPEOF(pborder) != REALSXP))
    error("invalid 'border' argument (must be numeric)");
  rset = VECTOR_ELT(ppats, 0);  /* get and check first pattern */
  if ((TYPEOF(rset) != VECSXP) || (length(rset) < 2))
    error("invalid 'pats' argument (wrong pattern type)");
  supp = VECTOR_ELT(rset, 1);   /* get support of first pattern */
  t    = TYPEOF(supp);          /* get and check the support type */
  if (((t != INTSXP) && (t != REALSXP)) || (length(supp) <= 0))
    error("invalid 'pats' argument (wrong support type)");
  k = 1;                        /* init. maximum pattern size */
  for (i = 0; i < n; i++) {     /* traverse the patterns */
    rset = VECTOR_ELT(ppats, (R_xlen_t)i);
    if ((length(rset) < 2) || (TYPEOF(rset) != VECSXP))
      error("invalid 'pats' argument (wrong pattern type)");
    items = VECTOR_ELT(rset,0); /* get and check the item set */
    if (TYPEOF(items) != INTSXP)
      error("invalid 'pats' argument (wrong item type)");
    supp  = VECTOR_ELT(rset,1); /* get and check the support value */
    if (length(supp) <= 0) error("invalid 'pats' argument");
    if ((TYPEOF(supp) != t) || (length(supp) <= 0))
      error("invalid 'pats' argument (wrong support type)");
    z = (size_t)length(items);  /* get the pattern size and */
    if (z > k) k = z;           /* update the maximal size */
  }
  /* n: number of patterns (item set/support pairs) */
  /* k: maximal size of a pattern (number of items) */

  /* --- create pattern set --- */
  patset = psr_create(n, (ITEM)k, 0, NULL);
  if (!patset) ERR_MEM();       /* create a pattern set */
  if ((pborder != R_NilValue) && (length(pborder) > 0)) {
    z = (size_t)length(pborder);/* get the border length */
    if (z > k+1) z = k+1;       /* limit border to pattern size */
    if      (TYPEOF(pborder) == INTSXP) {
      b = INTEGER(pborder);     /* get the support values */
      for (i = 2; i < z; i++) { /* traverse the pattern sizes */
        s = (b[i] < 0)    ? RSUPP_MAX : (RSUPP)b[i];
        psr_setbdr(patset, (ITEM)i, s);
      } }                       /* store support values in border */
    else if (TYPEOF(pborder) == REALSXP) {
      d = REAL(pborder);        /* get the support values */
      for (i = 2; i < z; i++) { /* traverse the pattern sizes */
        s = (IS_NA(d[i])) ? RSUPP_MAX : (RSUPP)d[i];
        psr_setbdr(patset, (ITEM)i, s);
      } }                       /* store support values in border */
    else error("invalid 'border' argument (must be numeric or NULL)");
  }                             /* check the border type */
  if (sig_aborted()) { sig_abort(0); psr_delete(patset,1); ERR_ABORT();}

  /* --- collect patterns --- */
  for (i = 0; i < n; i++) {     /* collect the item sets */
    rset  = VECTOR_ELT(ppats, (R_xlen_t)i);
    items = VECTOR_ELT(rset,0); /* get the items in the set */
    supp  = VECTOR_ELT(rset,1); /* and the support value */
    z = (size_t)length(items);  /* get the item array and its size */
    s = (t == INTSXP) ? (RSUPP)INTEGER(supp)[0] : (RSUPP)REAL(supp)[0];
    psr_addpat(patset, INTEGER(items), (ITEM)z, s, (void*)rset);
  }                             /* store pattern in pattern set */
  if (sig_aborted()) { sig_abort(0); psr_delete(patset,0); ERR_ABORT();}

  /* --- pattern set reduction --- */
  k     = psr_reduce(patset, method, addis);
  ppats = PROTECT(allocVector(VECSXP, (R_xlen_t)k));
  for (i = k = 0; i < n; i++) { /* traverse the reduced patterns */
    rset = (SEXP)psr_getorig(patset, i);
    if (rset) { SET_VECTOR_ELT(ppats, (R_xlen_t)k, rset); k++; }
  }                             /* collect the kept patterns */
  psr_delete(patset, 0);        /* clean up allocated memory */
  UNPROTECT(1);                 /* release the new pattern list */
  return ppats;                 /* return modified pattern set */
}  /* c4r_patred() */

/*--------------------------------------------------------------------*/

void R_init_fim4r (DllInfo *info)
{                               /* --- initialize module */
  R_registerRoutines(info, NULL, NULL, NULL, NULL);
  R_useDynamicSymbols(info, TRUE);
}  /* R_init_fim4r() */
