#-----------------------------------------------------------------------
# File    : Makefile
# Contents: build shared object for CoCoNAD in R
# Author  : Christian Borgelt
# History : 2015.08.06 file created
#           2016.04.20 creation of dependency files added
#-----------------------------------------------------------------------
SHELL    = /bin/bash
UTILDIR  = ../../util/src
MATHDIR  = ../../math/src
TRACTDIR = ../../tract/src
COCODIR  = ../../coconad/src

CC       = gcc -std=c99
# CC       = g++
RINC     = -I/usr/share/R/include
INCS     = -I$(UTILDIR) -I$(MATHDIR) -I$(TRACTDIR) -I$(COCODIR)
CFBASE   = -Wall -Wextra -Wconversion -Wno-unused-parameter \
           -Wimplicit-fallthrough=0 -pedantic -c -fPIC $(ADDFLAGS)
CFLAGS   = $(CFBASE) -DNDEBUG -DQUIET -O3 -funroll-loops
# CFLAGS   = $(CFBASE) -g
# CFLAGS   = $(CFBASE) -g -DSTORAGE

LD       = gcc
LDFLAGS  = $(ADDFLAGS)
LIBS     = -lm

HDRS     = $(UTILDIR)/fntypes.h   $(UTILDIR)/arrays.h   \
           $(UTILDIR)/memsys.h    $(UTILDIR)/symtab.h   \
           $(UTILDIR)/sigint.h    $(TRACTDIR)/tract.h   \
           $(TRACTDIR)/train.h    $(TRACTDIR)/patspec.h \
           $(TRACTDIR)/clomax.h   $(TRACTDIR)/cm4seqs.h \
           $(TRACTDIR)/cmfilter.h $(TRACTDIR)/report.h
OBJS     = arrays.o memsys.o idmap.o random.o sigint.o gamma.o \
           tract.o train.o patspec.o clomax.o report.o patred.o \
           cm4seqs.o cmfilter.o coconad.o ccnpsp.o coco4r.o $(ADDOBJS)

#-----------------------------------------------------------------------
# Build Shared Library
#-----------------------------------------------------------------------
all:          coco4r.so

coco4r.so:    $(OBJS) Makefile
	R CMD SHLIB -o coco4r.so $(OBJS)

#-----------------------------------------------------------------------
# Array Operations
#-----------------------------------------------------------------------
arrays.o:     $(UTILDIR)/arrays.h $(UTILDIR)/fntypes.h
arrays.o:     $(UTILDIR)/arrays.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)/arrays.c -o $@

#-----------------------------------------------------------------------
# Memory Management System for Objects of Equal Size
#-----------------------------------------------------------------------
memsys.o:     $(UTILDIR)/memsys.h
memsys.o:     $(UTILDIR)/memsys.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)/memsys.c -o $@

#-----------------------------------------------------------------------
# Symbol Table Management
#-----------------------------------------------------------------------
idmap.o:      $(UTILDIR)/symtab.h $(UTILDIR)/fntypes.h \
              $(UTILDIR)/arrays.h
idmap.o:      $(UTILDIR)/symtab.c Makefile
	$(CC) $(CFLAGS) $(INCS) -DIDMAPFN $(UTILDIR)/symtab.c -o $@

#-----------------------------------------------------------------------
# Random Number Generator Management
#-----------------------------------------------------------------------
random.o:     $(UTILDIR)/random.h
random.o:     $(UTILDIR)/random.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)/random.c -o $@

#-----------------------------------------------------------------------
# Interrupt Signal Handling
#-----------------------------------------------------------------------
sigint.o:     $(UTILDIR)/sigint.h
sigint.o:     $(UTILDIR)/sigint.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(UTILDIR)/sigint.c -o $@

#-----------------------------------------------------------------------
# Gamma Function
#-----------------------------------------------------------------------
gamma.o:      $(MATHDIR)/gamma.h
gamma.o:      $(MATHDIR)/gamma.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(MATHDIR)/gamma.c -o $@

#-----------------------------------------------------------------------
# Item and Transaction Management
#-----------------------------------------------------------------------
tract.o:      $(UTILDIR)/fntypes.h   $(UTILDIR)/arrays.h \
              $(UTILDIR)/symtab.h    $(TRACTDIR)/tract.h
tract.o:      $(TRACTDIR)/tract.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)/tract.c -o $@

#-----------------------------------------------------------------------
# (Spike) Train Management
#-----------------------------------------------------------------------
train.o:      $(UTILDIR)/fntypes.h   $(UTILDIR)/arrays.h \
              $(UTILDIR)/symtab.h    $(TRACTDIR)/tract.h \
              $(TRACTDIR)/train.h
train.o:      $(TRACTDIR)/train.c Makefile
	$(CC) $(CFLAGS) $(INCS) -DTRN_SURR $(TRACTDIR)/train.c -o $@

#-----------------------------------------------------------------------
# Item Set Reporter Management
#-----------------------------------------------------------------------
patspec.o:    $(UTILDIR)/fntypes.h  $(UTILDIR)/arrays.h \
              $(UTILDIR)/symtab.h   $(UTILDIR)/random.h \
              $(MATHDIR)/gamma.h    $(TRACTDIR)/tract.h \
              $(TRACTDIR)/patspec.h
patspec.o:    $(TRACTDIR)/patspec.c Makefile
	$(CC) $(CFLAGS) $(INCS) -DPSP_ESTIM -DPSP_TRAIN \
              $(TRACTDIR)/patspec.c -o $@

clomax.o:     $(UTILDIR)/fntypes.h  $(UTILDIR)/arrays.h \
              $(UTILDIR)/memsys.h   $(UTILDIR)/symtab.h \
              $(TRACTDIR)/tract.h   $(TRACTDIR)/clomax.h
clomax.o:     $(TRACTDIR)/clomax.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)/clomax.c -o $@

report.o:     $(UTILDIR)/fntypes.h  $(UTILDIR)/arrays.h \
              $(UTILDIR)/symtab.h   $(TRACTDIR)/tract.h \
              $(TRACTDIR)/patspec.h $(TRACTDIR)/report.h
report.o:     $(TRACTDIR)/report.c Makefile
	$(CC) $(CFLAGS) $(INCS) -DISR_PATSPEC -DISR_CLOMAX \
              -DISR_NONAMES -DISR_NOSTDFILE $(TRACTDIR)/report.c -o $@

#-----------------------------------------------------------------------
# Prefix Tree Management for Closed and Maximal Sequences
#-----------------------------------------------------------------------
cm4seqs.o:    $(UTILDIR)/fntypes.h   $(UTILDIR)/arrays.h  \
              $(UTILDIR)/memsys.h    $(UTILDIR)/symtab.h  \
              $(TRACTDIR)/tract.h    $(TRACTDIR)/report.h \
              $(TRACTDIR)/cm4seqs.h
cm4seqs.o:    $(TRACTDIR)/cm4seqs.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)/cm4seqs.c -o $@

#-----------------------------------------------------------------------
# List/Array Closed/Maximal Filter Management
#-----------------------------------------------------------------------
cmfilter.o:   $(UTILDIR)/fntypes.h   $(UTILDIR)/arrays.h \
              $(UTILDIR)/symtab.h    $(TRACTDIR)/tract.h \
              $(TRACTDIR)/report.h   $(TRACTDIR)/cmfilter.h
cmfilter.o:   $(TRACTDIR)/cmfilter.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)/cmfilter.c -o $@

#-----------------------------------------------------------------------
# Pattern Set Reduction Functions
#-----------------------------------------------------------------------
patred.o:     $(UTILDIR)/fntypes.h   $(UTILDIR)/arrays.h \
              $(UTILDIR)/symtab.h    $(TRACTDIR)/tract.h \
              $(TRACTDIR)/report.h   $(TRACTDIR)/patred.h
patred.o:     $(TRACTDIR)/patred.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(TRACTDIR)/patred.c -o $@

#-----------------------------------------------------------------------
# CoCoNAD
#-----------------------------------------------------------------------
coconad.o:    $(HDRS) $(COCODIR)/coconad.h
coconad.o:    $(COCODIR)/coconad.c Makefile
	$(CC) $(CFLAGS) $(INCS) -DCCN_ABORT $(COCODIR)/coconad.c -o $@

coconad.d:    $(COCODIR)/coconad.c
	$(CC) -MM $(CFLAGS) $(INCS) -DCCN_ABORT \
              $(COCODIR)/coconad.c > coconad.d

#-----------------------------------------------------------------------
# Pattern Spectrum Functions
#-----------------------------------------------------------------------
ccnpsp.o:     $(HDRS) $(UTILDIR)/random.h $(COCODIR)/coconad.h \
              $(COCODIR)/ccnpsp.h
ccnpsp.o:     $(COCODIR)/ccnpsp.c Makefile
	$(CC) $(CFLAGS) $(INCS) -DCCN_ABORT $(COCODIR)/ccnpsp.c -o $@

ccnpsp.d:     $(COCODIR)/ccnpsp.c
	$(CC) -MM $(CFLAGS) $(INCS) -DCCN_ABORT \
              $(COCODIR)/ccnpsp.c > ccnpsp.d

#-----------------------------------------------------------------------
# R Stuff
#-----------------------------------------------------------------------
coco4r.o:     coco4r.c Makefile
	$(CC) $(CFLAGS) $(INCS) $(RINC) coco4r.c -o $@

#-----------------------------------------------------------------------
# Source Distribution Packages
#-----------------------------------------------------------------------
dist:
	$(MAKE) clean
	mv Makefile local.mk
	mv dist.mk Makefile
	mkdir coconad; cp -r ../../coconad/{src,ex,doc} coconad; \
        mkdir -p tract/src; \
        cp ../../tract/src/{tract.[ch],train.[ch],patspec.[ch]} \
           ../../tract/src/{clomax.[ch],cm4seqs.[ch],cmfilter.[ch]} \
           ../../tract/src/{report.[ch],patred.[ch]} \
           ../../tract/src/{makefile,tract.mak} tract/src; \
        cp -r ../../tract/doc tract; \
        mkdir -p math/src; \
        cp ../../math/src/{gamma.[ch],makefile,math.mak} math/src; \
        cp -r ../../math/doc math; \
        mkdir -p util/src; \
        cp -r ../../util/src/{fntypes.h,arrays.[ch],symtab.[ch]} \
              ../../util/src/{memsys.[ch],random.[ch],sigint.[ch]} \
              ../../util/src/{makefile,util.mak} util/src; \
        cp -r ../../util/doc util
	cd ../..; rm -f coco4r*.tar.gz; \
        R CMD build coco4r > /dev/null 2>&1
	rm -rf coconad tract math util
	mv Makefile dist.mk
	mv local.mk Makefile
	cd ../..; rm -rf coco4r.Rcheck

#-----------------------------------------------------------------------
# Install
#-----------------------------------------------------------------------
install:
	cd ../..; mkdir rtmp; \
        R CMD INSTALL --build -l rtmp coco4r; \
        rm -rf rtmp; \
	mv coco4r_*_R_*.tar.gz ../dist/lnx

#-----------------------------------------------------------------------
# Clean up
#-----------------------------------------------------------------------
clean:
	-rm -f *.d *.o *.so *~ *.flc *.log core symbols.rds
